##############################################################################
#
# Copyright (c) 2005 Jim Washington and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

# minjson.py
# use python's parser to read minimal javascript objects.
# str's objects and fixes the text to write javascript.

# Thanks to Patrick Logan for starting the json-py project and making so many
# good test cases.

# Jim Washington 15 Jun 2005.

from re import compile, sub, search, DOTALL

# set to true if transmission size is more important than speed
# only affects writing.
alwaysStripWhiteSpace = False

# add to this string if you wish to exclude other math operators 
# from reading.
badOperators = '*'

#################################
#      read JSON object         #
#################################

slashstarcomment = compile(r'/\*.*?\*/',DOTALL)
doubleslashcomment = compile(r'//.*\n')

def _Read(aString):
        """Use eval in a 'safe' way to turn javascript expression into
           a python expression.  Allow only True, False, and None in global
           __builtins__, and since those map as true, false, null in 
           javascript, pass those as locals
        """
        try:
            result = eval(aString,
            {"__builtins__":{'True':True,'False':False,'None':None}},
            {'null':None,'true':True,'false':False})
        except NameError:
            raise ReadException, \
            "Strings must be quoted. Could not read '%s'." % aString
        except SyntaxError:
            raise ReadException, \
            "Syntax error.  Could not read '%s'." % aString
        return result

# badOperators is defined at the top of the module

# generate the regexes for math detection
rx = {}
for k in badOperators:
    if k in '+*':
        rx[k,'numeric operation'] = compile(r"\d*\s*\%s|\%s\s*\d*" % (k, k))
        rx[k,'location'] = compile(r"\%s" % (k))
    else:
        rx[k,'numeric operation'] = compile(r"\d*\s*%s|%s\s*\d*" % (k, k))
        rx[k,'location'] = compile("%s" % (k))

def _getStringState(aString):
    """return the list of required quote closures if the end of aString needs them
    to close quotes.
    """
    state = []
    for k in aString:
        if k in ['"',"'"]:
            if state and k == state[-1]:
                state.pop()
            else:
                state.append(k)
    return state

def _sanityCheckMath(aString):
    """just need to check that, if there is a math operator in the 
       client's JSON, it is inside a quoted string. This is mainly to 
       keep client from successfully sending 'D0S'*9**9**9**9...
       Return True if OK, False otherwise
    """
    for k in badOperators:
        #first check, is it a possible math operation?
        if rx[(k,'numeric operation')].search(aString) is not None:
            # OK.  possible math operation. get the operator's locations
            getlocs = rx[(k,'location')].finditer(aString)
            locs = [item.span()[0] for item in getlocs]
            #fortunately, this will be rare
            for idx in locs:
                # We only need to know the char is within open quote 
                # status.
                if not _getStringState(aString[:idx]):
                    return False
    return True

def safeRead(aString):
    """turn the js into happier python and check for bad operations
       before sending it to the interpreter
    """
    # strip leading and trailing whitespace; 
    aString = aString.strip()
    # get rid of trailing null if still there.
    if aString.endswith(chr(0)):
        aString = aString[:-1]
    # zap /* ... */ comments
    aString = slashstarcomment.sub('',aString)
    # zap // comments
    aString = doubleslashcomment.sub('',aString)
    # here, we only check for the * operator as a dos problem by default; 
    # additional operators may be excluded by editing badOperators
    # at the top of the module
    if _sanityCheckMath(aString):
        return _Read(aString)
    else:
        raise ReadException, 'Unacceptable JSON expression: %s' % aString

read = safeRead

#################################
#   write object as JSON        #
#################################

#alwaysStripWhiteSpace is defined at the top of the module

tfnTuple = (('True','true'),('False','false'),('None','null'),)

#whitespace = re.compile(r'\s')

def _replaceTrueFalseNone(aString):
    """replace True, False, and None with javascript counterparts"""
    for k in tfnTuple:
        if k[0] in aString:
            aString = aString.replace(k[0],k[1])
    return aString

def _handleCode(subStr,stripWhiteSpace):
    """replace True, False, and None with javascript counterparts if 
       appropriate, and strip white space if requested
    """
    if 'e' in subStr:
        #True, False, and None have 'e' in them. :)
        subStr = (_replaceTrueFalseNone(subStr))
    if stripWhiteSpace:
        # re.sub might do a better job, but takes longer.  
        # Spaces are the majority of the whitespace, anyway...
        #subStr = whitespace.sub('',subStr)
        subStr = subStr.replace(' ','')
    return subStr

def _pyexpr2jsexpr(aString, stripWhiteSpace):
    """Take advantage of python's formatting of string representations of
    objects.  Python always uses "'" to delimit strings.  If we split
    on that delimiter, we have a list that alternates non-string text with 
    string text.  Since string text is already properly escaped, we
    only need to replace True, False, and None in non-string text.

    if stripWhiteSpace is True, remove spaces, etc from the non-string
    text.
    """
    SQuotestr = None
    _EscapedSingleQuote = r"\'"
    if _EscapedSingleQuote in aString:
        #replace escaped single quotes with a marker
        SQuotestr = SQuotestrbase = '|'
        SQcount = 1
        while SQuotestr in aString:
            #on the odd chance that the marker is already there
            SQcount += 1
            SQuotestr = SQuotestrbase * SQcount
        aString = aString.replace(_EscapedSingleQuote,SQuotestr)
        
    #split the string on the real single-quotes
    splitStr = aString.split("'")
    outList = []
    alt = True
    for subStr in splitStr:
        #if alt is True, non-string; do replacements
        if alt:
            subStr = _handleCode(subStr,stripWhiteSpace) 
        outList.append(subStr)
        alt = not alt
    result = "'".join(outList)
    if SQuotestr:
        #put the escaped single-quotes back
        result = result.replace(SQuotestr,_EscapedSingleQuote)
    return result

def write(obj, stripWhiteSpace=alwaysStripWhiteSpace):
    """Represent the object as a string or unicode.  Do a fix-up
    with pyexpr2jsexpr"""
    try:
        aString = unicode(obj)
    except UnicodeDecodeError:
        aString = str(obj)
    if isinstance(obj,basestring):
        return aString
    return _pyexpr2jsexpr(aString,stripWhiteSpace)

class ReadException(Exception):
    pass

class WriteException(Exception):
    pass
